package org.netkernel.http.transport;

import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.net.URI;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentLinkedQueue;

import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.eclipse.jetty.http.HttpCookie;
import org.eclipse.jetty.http.HttpHeader;
import org.eclipse.jetty.http.HttpMethod;
import org.eclipse.jetty.io.EofException;
import org.eclipse.jetty.server.Response;
import org.netkernel.container.IKernel;
import org.netkernel.http.rep.IHttpRedirect;
import org.netkernel.http.rep.XSSSafeCookie;
import org.netkernel.http.transport.representation.HTTPRequestResponseRepresentation;
import org.netkernel.http.transport.representation.NKWebSocket;
import org.netkernel.http.util.SecurityUtils;
import org.netkernel.layer0.bnf.BNFGrammarFactory;
import org.netkernel.layer0.bnf.IGrammarGroup;
import org.netkernel.layer0.meta.IIdentifierGrammar;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.meta.impl.EndpointMetaImpl;
import org.netkernel.layer0.meta.impl.SimpleSpaceElementsImpl;
import org.netkernel.layer0.meta.impl.SpaceElementsImpl;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.representation.IBinaryStreamRepresentation;
import org.netkernel.layer0.representation.IHDSNode;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.layer0.urii.ActiveURIEncoding;
import org.netkernel.layer0.urii.SimpleIdentifierImpl;
import org.netkernel.layer0.util.RegexRewriterRule;
import org.netkernel.layer0.util.TupleList;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLReadable;
import org.netkernel.module.standard.endpoint.IStandardEndpoint;
import org.netkernel.module.standard.endpoint.IStandardOverlay;
import org.netkernel.module.standard.endpoint.StandardAccessorImpl;
import org.netkernel.module.standard.endpoint.TransparentOverlayImpl;
import org.netkernel.module.standard.endpoint.ConfiguredOverlayImpl.IConfig;
import org.netkernel.request.IRequest;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.urii.IEndpoint;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.INetKernelException;
import org.netkernel.urii.ISpace;
import org.netkernel.urii.impl.NetKernelError;
import org.w3c.dom.Document;
import org.w3c.dom.Node;


public class HTTPBridgeOverlayConfig implements IConfig
{
	private final RegexRewriterRule[] mRules;
	private final ISpace mDelegateSpace;
	private final String mExceptionHandler;
	private final IEndpoint mEndpoint;
	private final String mElementId;
	private final EndpointMetaImpl mMeta;
	private String mPostParamEncoding="UTF-8";

	private HTTPRequestSpace2 mSpace;
	
	private ConcurrentLinkedQueue<HTTPRequestSpace2Wrapper> mWebSocketSpaces=new ConcurrentLinkedQueue<HTTPRequestSpace2Wrapper>();
	
	public static final String HTTP_CACHE_EXPIRES_DELTA_SECONDS="HTTP_CACHE_EXPIRES_DELTA_SECONDS";
	public static final String HTTPBRIDGE_TRANSPARENT_PASS_THROUGH="HTTPBRIDGE_TRANSPARENT_PASS_THROUGH";
	
	
	public HTTPBridgeOverlayConfig(ISpace aDelegateSpace,ISpaceMeta aDelegateMeta, INKFRequestContext aContext, IEndpoint aEndpoint) throws Exception
	{	mDelegateSpace=aDelegateSpace;
		mExceptionHandler=(String)aContext.getParamValue("exceptionHandler");
		mEndpoint=aEndpoint;
		
		IKernel k=aContext.getKernelContext().getKernel();
		mSpace= new HTTPRequestSpace2((HTTPBridgeOverlay)mEndpoint,k);
		mSpace.onCommissionSpace(k);
		
		Document config=null;
		if(aContext.getParamValue("config")!=null)
		{	config=(Document)aContext.source("param:config",Document.class);
		}
		IGrammarGroup grammarRoot;
		if(config!=null)
		{	grammarRoot = BNFGrammarFactory.createChoiceGroup(null);
			XMLReadable c=new XMLReadable(config);
			List<RegexRewriterRule> rules = new ArrayList<RegexRewriterRule>();
			for (Iterator<Node> i=c.getNodes("/config/rewrite").iterator(); i.hasNext(); )
			{	XMLReadable rewrite=new XMLReadable(i.next());
				String match=rewrite.getText("match");
				String to=rewrite.getText("to");
				RegexRewriterRule rule = new RegexRewriterRule(match,to);
				rules.add(rule);
				BNFGrammarFactory.createRegex(grammarRoot, match);
			}
			mRules = new RegexRewriterRule[rules.size()];
			rules.toArray(mRules);
			String encoding=c.getText("/config/defaultPostParamEncoding");
			if (encoding.length()>0)
			{	mPostParamEncoding=encoding;
			}
		}
		else
		{	mRules = new RegexRewriterRule[1];
			String regex="(htt(p|ps)://.*)";
			mRules[0]=new RegexRewriterRule(regex, "$1");
			grammarRoot=BNFGrammarFactory.createGroup(null);
			BNFGrammarFactory.createRegex(grammarRoot, regex);
		}
		
		//setup meta
		mElementId = Utils.generateUniqueEndpointIdentifier(mDelegateSpace, aEndpoint, "HTTPBridgeOverlayConfig");
		IIdentifierGrammar grammar=BNFGrammarFactory.createGrammar(grammarRoot, aContext.getKernelContext().getKernel().getLogger());
		mMeta = new EndpointMetaImpl("HTTPBridge",null,null,null,IRequest.VERB_SOURCE,grammar,EndpointMetaImpl.NO_ARGS,EndpointMetaImpl.NO_RESPONSES,EndpointMetaImpl.NO_EXCEPTIONS);
		mMeta.addAdditionalField("private", true);
	}
	
	protected RegexRewriterRule[] getRewriteRules()
	{	return mRules;		
	}
	
	String getDefaultPostParamEncoding()
	{	return mPostParamEncoding;
	}
	
	protected ISpace getDelegateSpace()
	{	return mDelegateSpace;		
	}
	
	public int getWSConnectionCount()
	{	return mWebSocketSpaces.size();
	
	}
	
	public void onDecommission()
	{	for(ISpace space: mWebSocketSpaces)
		{	try
			{	space.onDecommissionSpace();
			}
			catch(Exception e)
			{	e.printStackTrace();				
			}
		}
		mSpace.releaseReferences();
	}
	
	public void remove(HTTPRequestSpace2Wrapper aSpace)
	{	mWebSocketSpaces.remove(aSpace);
	}

	public void onRequest(String aID, INKFRequestContext context) throws Exception
	{		//Pass through mode
			if(context.getThisRequest().getHeaderValue(HTTPBRIDGE_TRANSPARENT_PASS_THROUGH)!=null)
			{	//Perform any inner request rewriting
				String innerID=context.getThisRequest().getIdentifier();
				if(mRules.length>0)
				{	for(RegexRewriterRule rule : mRules )
					{	if(rule.matches(innerID))
						{	innerID=rule.map(innerID);
						}
					}
				}
				//Prepare sub-request into wrapped space only - no HTTP bridge space.
				INKFRequest req=context.createRequest(innerID);
				req.setRepresentationClass(IBinaryStreamRepresentation.class);
				context.createResponseFrom(context.issueRequestForResponse(req));
				return;
			}
		
			//Normal mode
			HTTPRequestResponseRepresentation httpRepresentation = (HTTPRequestResponseRepresentation)context.getThisRequest().getPrimary();
			//Prepare a root request context.
			if(httpRepresentation==null)
			{	throw new Exception("HTTPTransport Fulcrum space is misconfigured and is not providing the HTTPRequestResponseRepresentation to the HTTPBridge");				
			}
			HttpServletRequest httpRequest = httpRepresentation.getHttpRequest();
			NKWebSocket ws=httpRepresentation.getWebSocket();

			HTTPRequestSpace2Wrapper httpRequestSpace2=new HTTPRequestSpace2Wrapper((HTTPBridgeOverlay)mEndpoint, mSpace, httpRepresentation,this,ws);
			//Commission the prepared space
			httpRequestSpace2.onCommissionSpace(context.getKernelContext().getKernel());

			
			//HTTPRequestSpace httpRequestSpace = new HTTPRequestSpace(httpRequest,this, ws);
			if(ws!=null)
			{	//WEBSOCKET MODE
				mWebSocketSpaces.add(httpRequestSpace2);				
			}
			else
			{	//NORMAL MODE
				//HttpServletRequest httpRequest = httpRepresentation.getHttpRequest();
				IRequestScopeLevel injectionScope = context.getKernelContext().getThisKernelRequest().getRequestScope();
				injectionScope = RequestScopeLevelImpl.createDurableScopeLevel(httpRequestSpace2, injectionScope);			
				injectionScope=RequestScopeLevelImpl.createNonDurableScopeLevel(mDelegateSpace, injectionScope);
				
				//Perform any inner request rewriting
				String innerID=context.getThisRequest().getIdentifier();
				innerID = Utils.decode(innerID);
				if(mRules.length>0)
				{	for(RegexRewriterRule rule : mRules )
					{	if(rule.matches(innerID))
						{	innerID=rule.map(innerID);
						}
					}
				}
				//Prepare sub-request into prepared context
				INKFRequest req=context.createRequest(innerID);
				req.setRequestScope(injectionScope);
				req.setRepresentationClass(IBinaryStreamRepresentation.class);
				
				HttpServletResponse httpResponse = httpRepresentation.getHttpResponse();
				try
				{	//Commission the prepared context
					//httpRequestSpace2.onCommissionSpace(context.getKernelContext().getKernel());
					////////////////////////
					//Issue inner sub-request
					////////////////////////
					INKFResponseReadOnly innerResponse = context.issueRequestForResponse(req);
					if(!httpRepresentation.needsHandling())
					{	//Response has already been written by the HTTP request handler timeout code.
						INKFResponse response=context.createResponseFrom(null);
						response.setExpiry(INKFResponse.EXPIRY_ALWAYS);
						return; 					
					}
					IBinaryStreamRepresentation innerRepresentation = (IBinaryStreamRepresentation)innerResponse.getRepresentation();
					httpRequestSpace2.applyResponseMetadata(innerResponse.getHeaders(),context);
					
					if(innerResponse.hasHeader(IHttpRedirect.HTTP_REDIRECT_METADATA_IDENTIFIER))
					{	//Handle Redirects
						IHttpRedirect redirect=(IHttpRedirect)innerResponse.getHeader(IHttpRedirect.HTTP_REDIRECT_METADATA_IDENTIFIER);
						httpResponse.setStatus(redirect.getType());
						httpResponse.setHeader(HttpHeader.LOCATION.asString(), redirect.getURL());
						if(redirect.getType()==IHttpRedirect.HTTP_REDIRECT_300_MULTIPLE_CHOICES)
						{	//Write choices...
							httpResponse.setContentType("text/plain");
							OutputStreamWriter w=new OutputStreamWriter(httpResponse.getOutputStream());
							for(String s: redirect.getChoices())
							{	w.write(s);
								w.write("\n");
							}
							w.close();
						}
						else if(innerRepresentation!=null)
						{	//Comply with HTTP1.1 spec and send short note as body if it is present.
							writeStream(innerRepresentation, httpResponse, innerResponse.getMimeType(), httpRepresentation.getHttpRequest(), httpRequestSpace2);					
						}
					}
					else
					{	////////////////////////////////////////
						//Process HTTP Response
						////////////////////////////////////////
						
						//HEADERS WRITE
						boolean writestream=writeHeaders(innerResponse, httpRequest, httpResponse, httpRequestSpace2);
						if(writestream)
						{	if (innerRepresentation==null)
							{	throw context.createFormattedException("EX_NULL_RESPONSE", "MSG_NULL_RESPONSE", null, null);
							}
							else if (!HttpMethod.HEAD.is(httpRequest.getMethod()))
							{	writeStream(innerRepresentation, httpResponse, innerResponse.getMimeType(), httpRequest, httpRequestSpace2);
							}
						}
					}
				}
				catch(EofException e)
				{	//Requestor has gone away so no point handling Exception.
					context.logRaw(INKFRequestContext.LEVEL_FINE, e.getMessage());
					context.logRaw(INKFRequestContext.LEVEL_FINER, e.toString());
				}
				catch (NKFException e)
				{	handleException(e,context,injectionScope,httpResponse,httpRequest,httpRequestSpace2);
				}
				catch (NetKernelError e)
				{	handleException(e,context,injectionScope,httpResponse,httpRequest,httpRequestSpace2);
				}
				catch (Exception e)
				{	INetKernelException ex = context.createFormattedException("EX_IN_HTTP_BRIDGE",null,null,e);
					handleException(ex,context,injectionScope,httpResponse,httpRequest,httpRequestSpace2);
				}
			}
			httpRequestSpace2.releaseReferences();
			INKFResponse response=context.createResponseFrom(null);
			response.setExpiry(INKFResponse.EXPIRY_ALWAYS);
	}
	
	private boolean writeHeaders(INKFResponseReadOnly innerResponse, HttpServletRequest httpRequest, HttpServletResponse httpResponse, HTTPRequestSpace2Wrapper httpRequestSpace) throws IOException, NKFException
	{
		boolean writestream=true;
		Map h=httpRequestSpace.getResponseHeaders();
		
		//Snapshot time and us this as the absolute value for all derivative time based headers
		long now=System.currentTimeMillis();
		
		//httpResponse.containsHeader(HttpHeaders.DATE);
		httpResponse.setDateHeader(HttpHeader.DATE.asString(), now);
		
		if (!isHeaderRemoved(HttpHeader.SERVER.asString(),h)) httpResponse.setHeader(HttpHeader.SERVER.asString(),  HTTPTransportEndpoint.getServerHeader() );
		
		//Process Cookies
		Map cookies=httpRequestSpace.getResponseCookies();
		Iterator i=cookies.keySet().iterator();
		while(i.hasNext())
		{		Cookie c=(Cookie)cookies.get(i.next());
				if(c instanceof XSSSafeCookie)
				{	((Response)httpResponse).addCookie(SecurityUtils.covertToHttpOnlyJettyCookie(c));
				}
				else
				{	httpResponse.addCookie(c);
				}
		}
		
		if(innerResponse.hasHeader(HTTP_CACHE_EXPIRES_DELTA_SECONDS))
		{	String vs=innerResponse.getHeader(HTTP_CACHE_EXPIRES_DELTA_SECONDS).toString();
			try
			{	httpResponse.setDateHeader(HttpHeader.EXPIRES.asString(), now+Long.parseLong(vs)*1000L);
			}
			catch(Exception e)
			{	e.printStackTrace();  //Wrong number format.				
			}
			httpResponse.setHeader(HttpHeader.CACHE_CONTROL.asString(), "public, max-age="+vs);
		}

		//Process Redirect
		if(httpRequestSpace.shouldRedirect())
		{	String redirectURIString=httpRequestSpace.getResponseRedirect().replace(" ", "%20");
			URI redirectURI=URI.create(redirectURIString);
			//String redirectURIString=URI.create(httpRequestSpace.getResponseRedirect()).toASCIIString();
			String scheme=redirectURI.getScheme();

			if (scheme!=null && !scheme.equals("http") && !scheme.equals("https"))
			{	//remove scheme
				redirectURIString=redirectURI.getSchemeSpecificPart();
			}
			else
			{	redirectURIString=redirectURI.toASCIIString();
			}
			httpResponse.sendRedirect(redirectURIString);
			writestream=false;
		}
		else
		{	//Process Response Code
			httpResponse.setStatus(httpRequestSpace.getResponseCode());
		}
		
		Object rep=innerResponse.getRepresentation();
		if (rep!=null && writestream && !isHeaderRemoved(HttpHeader.ETAG.asString(),h))
		{	//Etag Processing
			String method=httpRequest.getMethod();
			if((HttpMethod.GET.is(method) || HttpMethod.HEAD.is(method)) && !innerResponse.isExpired())
			{	String etag=httpRequest.getHeader(HttpHeader.IF_NONE_MATCH.asString());  //Doesn't matter if null
				String tagString;
				try
				{	tagString=(String)innerResponse.getHeader("httpResponse:/header/Etag");
				}
				catch(Exception e)
				{	tagString="\""+Integer.toHexString(rep.hashCode())+"\"";
				}
				//Write the Etag
				httpResponse.setHeader(HttpHeader.ETAG.asString(),tagString);
				if(tagString.equals(etag))
				{	//Not modified
					writestream=false;
					httpResponse.setStatus(HttpServletResponse.SC_NOT_MODIFIED);
				}
			}
			//Write Expiration Header
			if(innerResponse.isExpired()
					&& (HttpMethod.GET.is(method) || HttpMethod.HEAD.is(method))
					&& !isHeaderRemoved(HttpHeader.EXPIRES.asString(),h)
					&& !innerResponse.hasHeader(HTTP_CACHE_EXPIRES_DELTA_SECONDS)
					)
			{	httpResponse.setHeader(HttpHeader.EXPIRES.asString(), "-1");
			}
		}
		
		if(writestream && rep!=null && !HttpMethod.HEAD.is(httpRequest.getMethod()) && !isHeaderRemoved(HttpHeader.LAST_MODIFIED.asString(),h))
		{	httpResponse.setDateHeader(HttpHeader.LAST_MODIFIED.asString(), now);
		}
		
		//Process user specified headers
		i=h.keySet().iterator();
		while(i.hasNext())
		{	String name=(String)i.next();
			Object value=h.get(name);
			if (value instanceof Long)
			{	httpResponse.setDateHeader(name, (Long)value);
			}
			else if (value instanceof Integer)
			{	httpResponse.setIntHeader(name, (Integer)value);
			}
			else if (value instanceof String)
			{	if (((String)value).length()>0)
				{	httpResponse.setHeader(name, value.toString());
				}
			}
			else
			{	httpResponse.setHeader(name, value.toString());
			}
		}
		return writestream;
	}
	
	private void writeStream(IBinaryStreamRepresentation innerRepresentation, HttpServletResponse httpResponse, String mimeType, HttpServletRequest httpRequest, HTTPRequestSpace2Wrapper httpRequestSpace2) throws Exception
	{	
		Map headers=httpRequestSpace2.getResponseHeaders();
		String encoding=innerRepresentation.getEncoding();
		if(encoding!=null)
		{	int gp=encoding.indexOf("gzip");
			if(gp>=0)
			{	if (!isHeaderSet("Content-Encoding",headers)) httpResponse.setHeader("Content-Encoding", "gzip");
				if(gp>0)
				{	encoding=encoding.substring(0,gp-1);  	//Strip +gzip tag							
				}
				else
				{	encoding=null;  											//Prevent writing gzip encoding on mimetype header.							
				}
			}
		}
		if (mimeType!=null)
		{	//Special case to make XHTML work with crappy Internet Explorer
			if(mimeType.equals("application/xhtml+xml"))
			{	if(httpRequest.getHeader(HttpHeader.USER_AGENT.asString()).contains("MSIE"))
				{	mimeType="text/html";
				}
			}
			if( (encoding!=null) && mimeType.startsWith("text/") )    //Encoding when text but we need a complete mimetype - encoding database see: http://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.7
			{	mimeType+=";charset="+encoding;
			}
			if (!isHeaderSet(HttpHeader.CONTENT_TYPE.asString(),headers))
			{	httpResponse.setContentType(mimeType);
			}
		}
		//Content-Length
		if (innerRepresentation instanceof IReadableBinaryStreamRepresentation && !isHeaderSet(HttpHeader.CONTENT_LENGTH.asString(),headers) )
		{	IReadableBinaryStreamRepresentation rbs = (IReadableBinaryStreamRepresentation)innerRepresentation;
			httpResponse.setContentLength(rbs.getContentLength());
		}
		
		OutputStream s=httpResponse.getOutputStream();
		innerRepresentation.write(s);
		//s.close();
	}
	
	private static boolean isHeaderRemoved(String aName, Map aHeaders)
	{	Object value=aHeaders.get(aName);
		return value instanceof String && ((String)value).length()==0;
	}

	private static boolean isHeaderSet(String aName, Map aHeaders)
	{	Object value=aHeaders.get(aName);
		return value!=null;
	}

	private void handleException(Throwable e, INKFRequestContext context, IRequestScopeLevel injectionContext, HttpServletResponse httpResponse, HttpServletRequest httpRequest, HTTPRequestSpace2Wrapper httpRequestSpace) throws Exception
	{	try
		{	httpResponse.reset();
		}
		catch(IllegalStateException ies)
		{	/*Can't reset the response - already committed*/
			context.logRaw(INKFRequestContext.LEVEL_FINE, "Exception but HTTPResponse is complete cannot be reset. "+e.getMessage());
			return;
		}
		try
		{
			//explicit serialisation so that handler considers exception as data
			INKFRequest req=context.createRequest("literal:exception");
			req.setVerb(INKFRequestReadOnly.VERB_TRANSREPRESENT);
			req.addPrimaryArgument(e);
			req.setRepresentationClass(IBinaryStreamRepresentation.class);
			req.setRequestScope(injectionContext);
			Object serializedException=context.issueRequest(req);
			
			if (mExceptionHandler!=null)
			{	req=context.createRequest(mExceptionHandler);
				req.setRequestScope(injectionContext);
				req.addArgumentByValue("operand",serializedException);
				req.setRepresentationClass(IBinaryStreamRepresentation.class);
				INKFResponseReadOnly<IBinaryStreamRepresentation> response = context.issueRequestForResponse(req);
				httpRequestSpace.applyResponseMetadata(response.getHeaders(),context);
				boolean writestream=writeHeaders(response, httpRequest, httpResponse, httpRequestSpace);
				if (writestream)
				{	writeStream(response.getRepresentation(), httpResponse, response.getMimeType(), httpRequest, httpRequestSpace);
				}
			}
			else
			{	httpResponse.setContentType("text/xml");
				OutputStream s=httpResponse.getOutputStream();
				IBinaryStreamRepresentation result = (IBinaryStreamRepresentation)serializedException;
				result.write(s);
				s.close();
			}
		}
		catch (Exception e2)
		{	context.logRaw(INKFRequestContext.LEVEL_WARNING, e2.toString());
			if (e instanceof INetKernelException)
			{	throw (INetKernelException)e;
			}
			else if (e instanceof NetKernelError)
			{	throw (NetKernelError)e;
			}
			else throw e2;
		}
	}

	public IMetaRepresentation getElementMeta(String aElementId)
	{	return mMeta;
	}

	public ISpaceElements getElements(INKFRequestContext aContext) 	throws Exception
	{	ISpaceElements result=new SimpleSpaceElementsImpl(new SimpleIdentifierImpl(mElementId),mEndpoint);
		return result;
	}

	public String getResolvedElement(INKFResolutionContext aContext) throws Exception
	{	String result=null;
		INKFRequestReadOnly toResolve=aContext.getRequestToResolve();
		if (toResolve.getVerb()==INKFRequestReadOnly.VERB_SOURCE)
		{	String id=toResolve.getIdentifier();
			for(RegexRewriterRule rule : mRules)
			{	if(rule.matches(id))
				{	result=mElementId;
					break;
				}
			}
		}
		return result;
	}

	public ISpace[] getSpaces(INKFRequestContext aContext) throws Exception
	{	return new ISpace[] { mSpace };
	}
	
}